import itertools

from fireant.queries import (
    DataSetQueryBuilder,
    DimensionChoicesQueryBuilder,
    DimensionLatestQueryBuilder,
)


class _Container(object):
    """
    This is a list of slicer elements, metrics or dimensions, used for accessing an element by key with a dot syntax.

    Example:

    .. code-block:: python

        slicer = Slicer(
            dimensions=[
                Dimension(key='my_dimension1')
            ]
        )
        slicer.dimensions.my_dimension1
    """
    def __init__(self, items):
        self._items = items
        for item in items:
            setattr(self, item.alias, item)

    def __iter__(self):
        return iter(self._items)

    def __getitem__(self, item):
        return getattr(self, item)

    def __contains__(self, item):
        return hasattr(self, item)

    def __eq__(self, other):
        """
        Checks if the other object is an instance of _Container and has the same number of items with matching keys.
        """
        return isinstance(other, _Container) and \
               all([a is not None
                    and b is not None
                    and a.alias == b.alias
                    for a, b in itertools.zip_longest(self._items, getattr(other, '_items', ()))])


class DataSet(object):
    """
    WRITEME
    """

    class Fields(_Container):
        pass

    def __init__(self, table, database, joins=(), fields=(),
                 hint_table=None,
                 always_query_all_metrics=False):
        """
        Constructor for a slicer.  Contains all the fields to initialize the slicer.

        :param table: (Required)
            A pypika Table reference. The primary table that this slicer will retrieve data from.

        :param database:  (Required)
            A Database reference. Holds the connection details used by this slicer to execute queries.

        :param fields: (Required: At least one)
            A list of fields mapping definitions of data in the data set. Fields are similar to a column in a database
            query result set. They are the values

        :param joins:  (Optional)
            A list of join descriptions for joining additional tables.  Joined tables are only used when querying a
            metric or dimension which requires it.

        :param hint_table: (Optional)
            A hint table used for querying dimension options.  If not present, the table will be used.  The hint_table
            must have the same definition as the table omitting dimensions which do not have a set of options (such as
            datetime or boolean dimensions) and the metrics.  This is provided to more efficiently query dimension
            options.

        :param always_query_all_metrics: (Default: False)
            When true, all metrics will be included in database queries in order to increase cache hits.
        """
        self.table = table
        self.database = database
        self.joins = joins

        self.hint_table = hint_table
        self.fields = DataSet.Fields(fields)

        # add query builder entry points
        self.query = DataSetQueryBuilder(self)
        self.latest = DimensionLatestQueryBuilder(self)

        self.always_query_all_metrics = always_query_all_metrics

        for field in fields:
            if not field.definition.is_aggregate:
                field.choices = DimensionChoicesQueryBuilder(self, field)

    def __eq__(self, other):
        return isinstance(other, DataSet) \
               and self.fields == other.fields

    def __repr__(self):
        return 'Slicer(fields=[{}])' \
            .format(','.join([repr(f)
                              for f in self.fields]))



from fireant.dataset.intervals import DatetimeInterval
from fireant.utils import alias_selector


def make_term_for_metrics(metric):
    return metric.definition.as_(alias_selector(metric.alias))


def make_term_for_dimension(dimension, window=None):
    """
    Makes a list of pypika terms for a given slicer definition.

    :param dimension:
        A slicer dimension.
    :param window:
        A window function to apply to the dimension definition if it is a continuous dimension.
    :return:
        a list of terms required to select and group by in a SQL query given a slicer dimension. This list will contain
        either one or two elements. A second element will be included if the dimension has a definition for its display
        field.
    """
    f_alias = alias_selector(dimension.alias)

    if window and isinstance(dimension, DatetimeInterval):
        return window(dimension.definition, dimension.interval_key).as_(f_alias)

    return dimension.definition.as_(f_alias)


def make_orders_for_dimensions(dimensions):
    """
    Creates a list of ordering for a slicer query based on a list of dimensions. The dimensions's display definition is
    used preferably as the ordering term but the definition is used for dimensions that do not have a display
    definition.

    :param dimensions:
    :return:
        a list of tuple pairs like (term, orientation) for ordering a SQL query where the first element is the term
        to order by and the second is the orientation of the ordering, ASC or DESC.
    """

    # Use the same function to make the definition terms to force it to be consistent.
    # Always take the last element in order to prefer the display definition.
    definitions = [make_term_for_dimension(dimension)
                   for dimension in dimensions]

    return [(definition, None)
            for definition in definitions]



from collections import Iterable
from enum import Enum
from functools import wraps

from fireant.utils import (
    immutable,
)
from .filters import (
    AntiPatternFilter,
    BooleanFilter,
    ComparatorFilter,
    ContainsFilter,
    ExcludesFilter,
    PatternFilter,
    RangeFilter,
)


class DataType(Enum):
    date = 1
    text = 2
    number = 3
    boolean = 4

    def __repr__(self):
        return self.name


CONTINUOUS_TYPES = [DataType.number, DataType.date]
DISCRETE_TYPES = [DataType.text, DataType.boolean]


class DataSetFilterException(Exception):
    def __init__(self, msg, allowed):
        super().__init__(msg)
        self.allowed = allowed


def restrict_types(allowed):
    def wrapper(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            if self.data_type not in allowed:
                msg = '`{func}` filter can only be used on fields with the following data types: {types}.' \
                    .format(func=func.__name__,
                            types=", ".join(map(repr, allowed)))
                raise DataSetFilterException(msg, allowed)
            return func(self, *args, **kwargs)

        return inner

    return wrapper


class Field:
    """
    The `Field` class represents a field in the `DataSet` object. A field is a mapping to a column in a database query.

    :param alias:
        A unique identifier used to identify the metric when writing slicer queries. This value must be unique over
        the metrics in the slicer.

    :param definition:
        A pypika expression which is used to select the value when building SQL queries. For metrics, this query
        **must** be aggregated, since queries always use a ``GROUP BY`` clause an metrics are not used as a group.

    :param data_type: {Number, Text, Boolean, Date}
        When True, the field's definition should be treated as an aggregate expression.

    :param label: (optional)
        A display value used for the metric. This is used for rendering the labels within the visualizations. If not
        set, the alias will be used as the default.

    :param is_aggregate: (default: False)
        When True, the field's definition should be treated as an aggregate expression.

    :param precision: (optional)
        A precision value for rounding decimals. By default, no rounding will be applied.

    :param prefix: (optional)
        A prefix for rendering labels in visualizations such as '$'

    :param suffix:
        A suffix for rendering labels in visualizations such as '€'
    """

    def __init__(self,
                 alias,
                 definition,
                 data_type: DataType = DataType.number,
                 label=None,
                 prefix: str = None,
                 suffix: str = None,
                 thousands: str = None,
                 precision: int = None,
                 hyperlink_template: str = None):
        self.alias = alias
        self.data_type = data_type
        self.definition = definition
        self.label = label \
            if label is not None \
            else alias
        self.prefix = prefix
        self.suffix = suffix
        self.thousands = thousands
        self.precision = precision
        self.hyperlink_template = hyperlink_template

    def eq(self, other):
        return ComparatorFilter(self.alias, self.definition, ComparatorFilter.Operator.eq, other)

    def ne(self, other):
        return ComparatorFilter(self.alias, self.definition, ComparatorFilter.Operator.ne, other)

    @restrict_types(CONTINUOUS_TYPES)
    def gt(self, other):
        return ComparatorFilter(self.alias, self.definition, ComparatorFilter.Operator.gt, other)

    @restrict_types(CONTINUOUS_TYPES)
    def ge(self, other):
        return ComparatorFilter(self.alias, self.definition, ComparatorFilter.Operator.gte, other)

    @restrict_types(CONTINUOUS_TYPES)
    def lt(self, other):
        return ComparatorFilter(self.alias, self.definition, ComparatorFilter.Operator.lt, other)

    @restrict_types(CONTINUOUS_TYPES)
    def le(self, other):
        return ComparatorFilter(self.alias, self.definition, ComparatorFilter.Operator.lte, other)

    @restrict_types(CONTINUOUS_TYPES)
    def between(self, lower, upper):
        """
        Creates a filter to filter a slicer query.

        :param lower:
            The start time of the filter. This is the beginning of the window for which results should be included.
        :param upper:
            The stop time of the filter. This is the end of the window for which results should be included.
        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is between the values
            start and stop.
        """
        return RangeFilter(self.alias, self.definition, lower, upper)

    def isin(self, values: Iterable):
        """
        Creates a filter that filters to rows where

        :param values:
            An iterable of value to constrain the slicer query results by.

        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is one of a set of
            values. Opposite of #notin.
        """
        return ContainsFilter(self.alias, self.definition, values)

    def notin(self, values):
        """
        Creates a filter to filter a slicer query.

        :param values:
            An iterable of value to constrain the slicer query results by.

        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is *not* one of a set of
            values. Opposite of #isin.
        """
        return ExcludesFilter(self.alias, self.definition, values)

    @restrict_types([DataType.text])
    def like(self, pattern, *patterns):
        return PatternFilter(self.alias, self.definition, pattern, *patterns)

    @restrict_types([DataType.text])
    def not_like(self, pattern, *patterns):
        return AntiPatternFilter(self.alias, self.definition, pattern, *patterns)

    @restrict_types([DataType.boolean])
    def is_(self, value: bool):
        """
        Creates a filter to filter a slicer query.

        :param value:
            True or False
        :return:
            A slicer query filter used to filter a slicer query to results where this dimension is True or False.
        """
        return BooleanFilter(self.alias, self.definition, value)

    def __eq__(self, other):
        return self.eq(other)

    def __ne__(self, other):
        return self.ne(other)

    def __gt__(self, other):
        return self.gt(other)

    def __ge__(self, other):
        return self.ge(other)

    def __lt__(self, other):
        return self.lt(other)

    def __le__(self, other):
        return self.le(other)

    def __repr__(self):
        return "{alias}[{data_type}]".format(alias=self.alias,
                                             data_type=self.data_type)

    def __hash__(self):
        return hash(repr(self))

    @property
    @immutable
    def share(self):
        self._share = True
        return self



from .base import Database
from .mysql import MySQLDatabase
from .postgresql import PostgreSQLDatabase
from .redshift import RedshiftDatabase
from .snowflake import SnowflakeDatabase
from .vertica import VerticaDatabase



from functools import partial

from .modifiers import DimensionModifier


class NumericInterval(DimensionModifier):
    def __init__(self, dimension, size=1, offset=0):
        self.size = size
        self.offset = offset
        super().__init__(dimension)

    def __eq__(self, other):
        return all([isinstance(other, NumericInterval),
                    self.size == other.size,
                    self.offset == other.offset])

    def __hash__(self):
        return hash(repr(self))


DATETIME_INTERVALS = ('hour', 'day', 'week', 'month', 'quarter', 'year')


class DatetimeInterval(DimensionModifier):
    def __init__(self, dimension, interval_key):
        super().__init__(dimension)
        self.interval_key = interval_key

    def __eq__(self, other):
        return isinstance(other, DatetimeInterval) \
               and self.alias == other.alias

    def __repr__(self):
        wrapped_key = super().__getattribute__('wrapped_key')
        wrapped = super().__getattribute__(wrapped_key)
        return '{}({})'.format(self.interval_key, repr(wrapped))

    def __hash__(self):
        return hash(repr(self))


hour, day, week, month, quarter, year = [partial(DatetimeInterval, interval_key=key)
                                         for key in DATETIME_INTERVALS]



import copy
from functools import partial

from fireant.dataset.fields import Field
from pypika.terms import (
    ComplexCriterion,
    Criterion,
    Term,
)
from .field_helper import make_term_for_dimension


def adapt_for_reference_query(reference_parts, database, dimensions, metrics, filters, references):
    if reference_parts is None:
        return dimensions, metrics, filters

    ref_dimension, time_unit, interval = reference_parts

    ref_metrics = _make_reference_metrics(metrics,
                                          references[0].reference_type.alias)
    offset_func = partial(database.date_add,
                          date_part=time_unit,
                          interval=interval)
    ref_dimensions = _make_reference_dimensions(database,
                                                dimensions,
                                                ref_dimension,
                                                offset_func)
    ref_filters = _make_reference_filters(filters,
                                          ref_dimension,
                                          offset_func)
    return ref_dimensions, ref_metrics, ref_filters


def _make_reference_dimensions(database, dimensions, ref_dimension, offset_func):
    def replace_reference_dimension(dimension):
        ref_dimension_copy = copy.copy(dimension)
        if hasattr(ref_dimension_copy, 'dimension'):
            ref_dimension_copy.dimension = copy.copy(dimension.dimension)

        ref_definition = make_term_for_dimension(ref_dimension_copy, database.trunc_date)
        ref_dimension_copy.definition = offset_func(ref_definition)
        return ref_dimension_copy

    return [replace_reference_dimension(dimension)
            if dimension is ref_dimension
            else dimension
            for dimension in dimensions]


def _make_reference_metrics(metrics, ref_key):
    return [Field(metric.alias + '_' + ref_key,
                  metric.definition,
                  label=metric.label,
                  prefix=metric.prefix,
                  suffix=metric.suffix,
                  precision=metric.precision)
            for metric in metrics]


def _make_reference_filters(filters, ref_dimension, offset_func):
    """
    Copies and replaces the reference dimension's definition in all of the filters applied to a slicer query.

    This is used to shift the dimension filters to fit the reference window.

    :param filters:
    :param ref_dimension:
    :param offset_func:
    :return:
    """
    offset_ref_dimension_definition = offset_func(ref_dimension.definition)

    reference_filters = []
    for ref_filter in map(copy.deepcopy, filters):
        ref_filter.definition = _apply_to_term_in_criterion(ref_dimension.definition,
                                                            offset_ref_dimension_definition,
                                                            ref_filter.definition)
        reference_filters.append(ref_filter)

    return reference_filters


def _apply_to_term_in_criterion(target: Term,
                                replacement: Term,
                                criterion: Criterion):
    """
    Finds and replaces a term within a criterion.  This is necessary for adapting filters used in reference queries
    where the reference dimension must be offset by some value.  The target term is found inside the criterion and
    replaced with the replacement.

    :param target:
        The target term to replace in the criterion. It will be replaced in all locations within the criterion with
        the func applied to itself.
    :param replacement:
        The replacement for the term.
    :param criterion:
        The criterion to replace the term in.
    :return:
        A criterion identical to the original criterion arg except with the target term replaced by the replacement arg.
    """
    if isinstance(criterion, ComplexCriterion):
        criterion.left = _apply_to_term_in_criterion(target, replacement, criterion.left)
        criterion.right = _apply_to_term_in_criterion(target, replacement, criterion.right)
        return criterion

    for attr in ['term', 'left', 'right']:
        if hasattr(criterion, attr) and str(getattr(criterion, attr)) == str(target):
            setattr(criterion, attr, replacement)

    return criterion



import pandas as pd

from pypika import Order


def _get_window(limit, offset):
    start = offset
    end = offset + limit \
        if None not in (offset, limit) \
        else limit
    return start, end


def _apply_sorting(orders):
    sort_values, ascending = zip(*[(order[0].alias, order[1] is Order.asc)
                                   for order in orders])
    return list(sort_values), ascending


def paginate(data_frame, widgets, orders=(), limit=None, offset=None):
    """
    :param data_frame:
        The result set to paginate.
    :param widgets:
        An iterable of widgets that the pagination is being applied for.
    :param orders:
        An iterable of (<Dimension/Metric>, pypika.Order)
    :param limit:
        A limit of the number of data points/series
    :param offset:
        A offset of the number of data points/series
    :return:
        A paginated data frame. If the widget required grouped pagination, then there should be an upperbound
        `limit*(n_index_level_0)`. Otherwise the data frame should have the same length as the limit.
    """
    if len(data_frame) == 0:
        return data_frame

    start, end = _get_window(limit, offset)
    group_pagination = isinstance(data_frame.index, pd.MultiIndex) \
                       and any([getattr(widget, 'group_pagination', False)
                                for widget in widgets])

    if group_pagination:
        return _group_paginate(data_frame, start, end, orders)
    return _simple_paginate(data_frame, start, end, orders)


def _simple_paginate(data_frame, start=None, end=None, orders=()):
    """
    Applies pagination which limits the number of rows in the data frame.

    :param data_frame:
        A data frame to paginate
    :param start:
        The index starting point to slice the data frame at
    :param end:
        The index ending point to slice the data frame at
    :param orders:
        A list of tuples that contain a slicer field definition (with an alias matching the columns of the data frame)
        and a pypika.Order.
    """
    if orders:
        sort, ascending = _apply_sorting(orders)
        data_frame = data_frame.sort_values(by=sort,
                                            ascending=ascending)

    return data_frame[start:end]


def _index_isnull(data_frame):
    if isinstance(data_frame.index, pd.MultiIndex):
        return [any(pd.isnull(value)
                    for value in level)
                for level in list(data_frame.index)]

    return pd.isnull(data_frame.index)


def _group_paginate(data_frame, start=None, end=None, orders=()):
    """
    Applies pagination which limits the number of rows in the data frame grouped by the zeroth index level. This will
    in turn paginate the number of series in the data frame.

    :param data_frame:
        A data frame to paginate
    :param start:
        The index starting point to slice the data frame at
    :param end:
        The index ending point to slice the data frame at
    :param orders:
        A list of tuples that contain a slicer field definition (with an alias matching the columns of the data frame)
        and a pypika.Order.
    """
    dimension_levels = data_frame.index.names[1:]
    dimension_groups = data_frame.groupby(level=dimension_levels)

    # Do not apply ordering on the 0th dimension !!!
    # This would not have any result since the X-Axis on a chart is ordered sequentially
    orders = [order
              for order in orders
              if order[0].alias != data_frame.index.names[0]]

    if orders:
        # FIXME this should aggregate according to field definition, instead of sum
        # Need a way to interpret definitions in python code in order to do that
        aggregated_df = dimension_groups.sum()

        sort, ascending = _apply_sorting(orders)
        sorted_df = aggregated_df.sort_values(by=sort,
                                              ascending=ascending)
        sorted_dimension_values = tuple(sorted_df.index)[start:end]

    else:
        sorted_dimension_values = tuple(dimension_groups.apply(lambda g: g.name))[start:end]

    sorted_dimension_values = pd.Index(sorted_dimension_values, name=dimension_levels[0]) \
        if len(dimension_levels) == 1 \
        else pd.MultiIndex.from_tuples(sorted_dimension_values, names=dimension_levels)

    def _apply_pagination(df):
        # This function applies sorting by using the sorted dimension values as an index to select values in the right
        # order out of the data frame. The index must be filtered to only values that are in this data frame, since
        # there might be missing combinations of index values.
        dfx = df.reset_index(level=0, drop=True)
        value_in_index = sorted_dimension_values.isin(dfx.index)
        index_slice = sorted_dimension_values[value_in_index]

        # Need to include nulls so append them to the end of the sorted data frame
        isnull = _index_isnull(dfx)

        return dfx.loc[index_slice, :].append(dfx[isnull])

    return data_frame \
        .sort_values(data_frame.index.names[0], ascending=True) \
        .groupby(level=0) \
        .apply(_apply_pagination)



from pypika import (
    Dialects,
    MySQLQuery,
    enums,
    functions as fn,
    terms,
)
from .base import Database


class Trunc(terms.Function):
    """
    Wrapper for a custom MySQL TRUNC function (installed via a custom FireAnt MySQL script)
    """

    def __init__(self, field, date_format, alias=None):
        super(Trunc, self).__init__('dashmore.TRUNC', field, date_format, alias=alias)
        # Setting the fields here means we can access the TRUNC args by name.
        self.field = field
        self.date_format = date_format
        self.alias = alias


class DateAdd(terms.Function):
    """
    Override for the MySQL specific DateAdd function which expects an interval instead of the date part and interval
    unit e.g. DATE_ADD("date", INTERVAL 1 YEAR)
    """

    def __init__(self, field, interval_term, alias=None):
        super(DateAdd, self).__init__('DATE_ADD', field, interval_term, alias=alias)


class MySQLDatabase(Database):
    """
    MySQL client that uses the PyMySQL module.
    """
    # The pypika query class to use for constructing queries
    query_cls = MySQLQuery

    def __init__(self, host='localhost', port=3306, database=None,
                 user=None, password=None, charset='utf8mb4', max_processes=1, cache_middleware=None):
        super(MySQLDatabase, self).__init__(host, port, database,
                                            max_processes=max_processes,
                                            cache_middleware=cache_middleware)
        self.user = user
        self.password = password
        self.charset = charset

    def _get_connection_class(self):
        # Nesting inside a function so the import does not cause issues if users have not installed the 'mysql' extra
        # when installing
        import pymysql

        class MySQLConnection(pymysql.connections.Connection):
            """
            PyMySQL has deprecated context managers in the connection class.
            To make the functionality consistent with other database drivers, we override
            the context manager to return the connection instead of the cursor object.

            This also fixes an issue where connections were not being closed in the current PyMySQL
            context manager implementation!
            https://github.com/PyMySQL/PyMySQL/issues/735
            """

            def __enter__(self):
                return self

            def __exit__(self, exc, value, traceback):
                self.close()

        return MySQLConnection

    def connect(self):
        """
        Returns a MySQL connection

        :return: pymysql Connection class
        """
        import pymysql
        connection_class = self._get_connection_class()
        return connection_class(host=self.host, port=self.port, db=self.database,
                                user=self.user, password=self.password,
                                charset=self.charset, cursorclass=pymysql.cursors.Cursor)

    def trunc_date(self, field, interval):
        return Trunc(field, str(interval))

    def to_char(self, definition):
        return fn.Cast(definition, enums.SqlTypes.CHAR)

    def date_add(self, field, date_part, interval):
        # adding an extra 's' as MySQL's interval doesn't work with 'year', 'week' etc, it expects a plural
        interval_term = terms.Interval(**{'{}s'.format(str(date_part)): interval, 'dialect': Dialects.MYSQL})
        return DateAdd(field, interval_term)



from pypika import JoinType


class Join(object):
    """
    WRITEME
    """

    def __init__(self, table, criterion, join_type=JoinType.inner):
        self.table = table
        self.criterion = criterion
        self.join_type = join_type

    def __repr__(self):
        return '{type} JOIN {table} ON {criterion}'.format(type=self.join_type,
                                                           table=self.table,
                                                           criterion=self.criterion)

    def __gt__(self, other):
        return self.table._table_name < other.table._table_name



from typing import (
    Iterable,
    Union,
)

from fireant import (
    Field,
    Operation,
    utils,
)
from .base import (
    MetricRequiredException,
    Widget,
)


class Series:
    type = None
    needs_marker = False
    stacking = None

    def __init__(self, metric: Union[Field, Operation], stacking=None):
        self.metric = metric
        self.stacking = self.stacking or stacking

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__,
                               repr(self.metric))


class ContinuousAxisSeries(Series):
    pass


class Axis:
    def __init__(self, series: Iterable[Series], label=None, y_axis_visible=True):
        self._series = series or []
        self.label = label
        self.y_axis_visible = y_axis_visible

    def __iter__(self):
        return iter(self._series)

    def __len__(self):
        return len(self._series)

    def __repr__(self):
        return "axis({})".format(", ".join(map(repr, self)))


class ChartWidget(Widget):
    class LineSeries(ContinuousAxisSeries):
        type = 'line'

    class AreaSeries(ContinuousAxisSeries):
        type = 'area'

    class AreaStackedSeries(AreaSeries):
        stacking = "normal"

    class AreaPercentageSeries(AreaSeries):
        stacking = "percent"

    class PieSeries(Series):
        type = 'pie'

    class BarSeries(Series):
        type = 'bar'

    class StackedBarSeries(BarSeries):
        stacking = "normal"

    class ColumnSeries(Series):
        type = 'column'

    class StackedColumnSeries(ColumnSeries):
        stacking = "normal"

    @utils.immutable
    def axis(self, *series: Series, **kwargs):
        """
        (Immutable) Adds an axis to the Chart.

        :param axis:
        :return:
        """

        self.items.append(Axis(series, **kwargs))

    @property
    def metrics(self):
        """
        :return:
            A set of metrics used in this chart. This collects all metrics across all axes.
        """
        if 0 == len(self.items):
            raise MetricRequiredException(str(self))

        seen = set()
        return [metric
                for axis in self.items
                for series in axis
                for metric in getattr(series.metric, 'metrics', [series.metric])
                if not (metric.alias in seen or seen.add(metric.alias))]

    @property
    def operations(self):
        return utils.ordered_distinct_list_by_attr([operation
                                                    for axis in self.items
                                                    for series in axis
                                                    if isinstance(series.metric, Operation)
                                                    for operation in [series.metric] + series.metric.operations])



import codecs
import os
import re

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
      name='fireant',
      version=find_version("fireant", "__init__.py"),

      author='KAYAK, GmbH',
      author_email='bandit@kayak.com',

      # License
      license='Apache License Version 2.0',

      packages=['fireant',
                'fireant.database',
                'fireant.dataset',
                'fireant.queries',
                'fireant.widgets'],

      # Include additional files into the package
      include_package_data=True,

      # Details
      url='https://github.com/kayak/fireant',

      description='Data analysis tool for Python and Jupyter Notebooks',
      long_description="Fireant is a tool which aims to simplify the process of accessing analytical data and building "
                       "dashboards and reports.  It eliminates the need for handwritten queries or code to process "
                       "data. "
                       "The main feature, the slicer, is a schema that is used to configure metrics, the data being "
                       "visualized, and dimensions, the spatiality over which metrics can be extended and grouped. "
                       "Included in the Slicer is a Slicer Manager API which exposes methods to render tables and "
                       "charts "
                       "in a variety of different formats.\n\n"

                       "Fireant is a great compliment to Jupter notebooks using matplotlib and pandas. It can be used "
                       "in "
                       "web projects with Highcharts and Datatables.js.  It can even just provide a layer of "
                       "abstraction "
                       "for accessing analytical data akin to ORM libraries in software projects.",

      classifiers=[
          'License :: OSI Approved :: Apache Software License',
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Programming Language :: PL/SQL',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Operating System :: POSIX',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
      ],
      keywords=('fireant python query builder querybuilder sql mysql postgres psql oracle vertica aggregated '
                'relational database rdbms business analytics bi data science analysis pandas'),

      install_requires=required,
      tests_require=[
          'mock'
      ],
      extras_require={
          'vertica': ['vertica-python==0.7.3'],
          'snowflake': ['snowflake-connector-python==1.7.2',
                        'cryptography==2.4.2'],
          'mysql': ['pymysql==0.8.0'],
          'redshift': ['psycopg2==2.7.3.2'],
          'postgresql': ['psycopg2==2.7.3.2'],
          'matplotlib': ['matplotlib'],
      },

      test_suite='fireant.tests',
)



import re

import numpy as np
import pandas as pd
from datetime import (
    date,
    datetime,
)

from fireant.dataset.fields import DataType
from fireant.dataset.totals import TOTALS_MARKERS
from fireant.utils import (
    filter_kwargs,
)

RAW_VALUE = 'raw'
INF_VALUE = "Inf"
NULL_VALUE = 'null'
BLANK_VALUE = ''
TOTALS_VALUE = '$totals'
TOTALS_LABEL = 'Totals'

epoch = np.datetime64(datetime.utcfromtimestamp(0))
milliseconds = np.timedelta64(1, 'ms')

DATE_FORMATS = {
    'iso': '%Y-%m-%dT%H:%M:%S',
    'hour': '%Y-%m-%d %H:00',
    'day': '%Y-%m-%d',
    'week': 'W%W %Y-%m-%d',
    'month': '%b %Y',
    'quarter': '%Y-%q',
    'year': '%Y',
}


@filter_kwargs
def date_as_string(value, interval_key=None):
    f = DATE_FORMATS.get(interval_key, '%Y-%m-%d')
    return value.strftime(f)


@filter_kwargs
def date_as_millis(value):
    if isinstance(value, date) and not isinstance(value, datetime):
        value = datetime.combine(value, datetime.min.time())
    return int(1000 * value.timestamp())


@filter_kwargs
def return_none(value):
    return None


@filter_kwargs
def _format_decimal(value, thousands=',', precision=None):
    if not isinstance(value, (int, float)):
        return value

    precision_pattern = '{{:{}.{}f}}'.format(thousands, precision) \
        if precision is not None \
        else '{{:{}f}}'.format(thousands)

    f_value = precision_pattern.format(value)
    if precision is None:
        f_value = f_value.rstrip('0').rstrip('.')
    return f_value


@filter_kwargs
def wrap_styling(value, prefix=None, suffix=None):
    if value is None:
        return value

    return '{prefix}{value}{suffix}' \
        .format(prefix=prefix or '',
                suffix=suffix or '',
                value=value)


@filter_kwargs
def _identity(value):
    return value


UNSAFE_CHARS = re.compile(r'[^\w\d\-:()]')


def json_value(value):
    """
    This function will return only values safe for JSON
    """
    if pd.isnull(value):
        return None
    if isinstance(value, float) and np.isinf(value):
        return 'inf'
    if value in TOTALS_MARKERS:
        return TOTALS_VALUE
    return value


def safe_value(value):
    if value is None or value is '' or pd.isnull(value):
        return NULL_VALUE
    if isinstance(value, float) and np.isinf(value):
        return 'inf'
    if value in TOTALS_MARKERS:
        return TOTALS_VALUE

    str_value = date_as_string(value, interval_key='iso') \
        if isinstance(value, date) \
        else str(value)
    return UNSAFE_CHARS.sub('$', str_value)


@filter_kwargs
def _format_date_field(value, date_as=date_as_string, **kwargs):
    return date_as(value, **kwargs)


RAW_FIELD_FORMATTER = {
    DataType.date: _format_date_field,
}


def raw_value(value, field, date_as=date_as_string):
    """
    Converts a raw metric value into a safe type. This will change dates into strings, NaNs into Nones, and np types
    into their corresponding python types.

    :param value:
        The raw metric value.
    :param field:
        The slicer field that the value represents.
    :param date_as:
    """
    if value is None or pd.isnull(value):
        return None
    if isinstance(value, float) and np.isinf(value):
        return INF_VALUE
    if value in TOTALS_MARKERS:
        return TOTALS_VALUE

    formatter = RAW_FIELD_FORMATTER.get(field.data_type, _identity)
    return formatter(value,
                     date_as=date_as,
                     interval_key='iso')


def _format_number_field_value(value, **kwargs):
    number = _format_decimal(value, **kwargs)
    return wrap_styling(number, **kwargs)


@filter_kwargs
def _format_boolean_field(value):
    return str(value).lower()


FIELD_DISPLAY_FORMATTER = {
    DataType.date: _format_date_field,
    DataType.number: _format_number_field_value,
    DataType.boolean: _format_boolean_field,
    DataType.text: return_none,
}


def display_value(value, field, date_as=date_as_string):
    """
    Converts a metric value into the display value by applying formatting.

    :param value:
        The raw metric value.
    :param field:
        The slicer field that the value represents.
    :param date_as:
        A format function for datetimes.
    :return:
        A formatted string containing the display value for the metric.
    """
    if pd.isnull(value):
        return BLANK_VALUE
    if isinstance(value, float) and np.isinf(value):
        return INF_VALUE
    if value in TOTALS_MARKERS:
        return TOTALS_LABEL

    format_kwargs = {key: getattr(field, key, None)
                     for key in ('prefix', 'suffix', 'thouands', 'precision', 'interval_key')}

    formatter = FIELD_DISPLAY_FORMATTER.get(field.data_type, _identity)
    return formatter(value,
                     date_as=date_as,
                     **format_kwargs)



import itertools
from datetime import timedelta

import pandas as pd
from fireant import (
    DataType,
    formats,
    utils,
)
from fireant.dataset.totals import TOTALS_MARKERS
from fireant.reference_helpers import (
    reference_alias,
    reference_label,
    reference_prefix,
    reference_suffix,
)

from .base import TransformableWidget
from .chart_base import (
    ChartWidget,
    ContinuousAxisSeries,
)

DEFAULT_COLORS = (
    "#DDDF0D",
    "#55BF3B",
    "#DF5353",
    "#7798BF",
    "#AAEEEE",
    "#FF0066",
    "#EEAAEE",
    "#DF5353",
    "#7798BF",
    "#AAEEEE",
)

DASH_STYLES = (
    'Solid',
    'Dash',
    'Dot',
    'DashDot',
    'LongDash',
    'LongDashDot',
    'ShortDash',
    'ShortDashDot',
    'LongDashDotDot',
    'ShortDashDotDot',
    'ShortDot',
)

MARKER_SYMBOLS = (
    "circle",
    "square",
    "diamond",
    "triangle",
    "triangle-down",
)

SERIES_NEEDING_MARKER = (ChartWidget.LineSeries, ChartWidget.AreaSeries)
TS_UPPER_BOUND = pd.Timestamp.max - timedelta(seconds=1)


class HighCharts(ChartWidget, TransformableWidget):
    # Pagination should be applied to groups of the 0th index level (the x-axis) in order to paginate series
    group_pagination = True

    def __init__(self, title=None, colors=None, x_axis_visible=True, tooltip_visible=True):
        super(HighCharts, self).__init__()
        self.title = title
        self.colors = colors or DEFAULT_COLORS
        self.x_axis_visible = x_axis_visible
        self.tooltip_visible = tooltip_visible

    def __repr__(self):
        return ".".join(["HighCharts()"] + [repr(axis) for axis in self.items])

    def transform(self, data_frame, slicer, dimensions, references):
        """
        - Main entry point -

        Transforms a data frame into HighCharts JSON format.

        See https://api.highcharts.com/highcharts/

        :param data_frame:
            The data frame containing the data. Index must match the dimensions parameter.
        :param slicer:
            The slicer that is in use.
        :param dimensions:
            A list of dimensions that are being rendered.
        :param references:
            A list of references that are being rendered.
        :return:
            A dict meant to be dumped as JSON.
        """
        colors = itertools.cycle(self.colors)

        is_timeseries = dimensions and dimensions[0].data_type == DataType.date

        dimension_map = {dimension_alias: dimension
                         for dimension_alias, dimension in zip(data_frame.index.names, dimensions)}
        num_dimensions = len(dimension_map)
        dimension_fields = [dimension_map[field_alias]
                            for field_alias in data_frame.index.names[:num_dimensions]]

        # Timestamp.max is used as a marker for rolled up dimensions (totals). Filter out the totals value for the
        # dimension used for the x-axis
        if is_timeseries and len(data_frame) > 0:
            data_frame = self._remove_date_totals(data_frame)

        # Group the results by index levels after the 0th, one for each series
        # This will result in a series for every combination of dimension values and each series will contain a data set
        # across the 0th dimension (used for the x-axis)
        series_data_frames = self._group_by_series(data_frame)

        total_num_series = sum([len(axis)
                                for axis in self.items])

        y_axes, series = [], []
        for axis_idx, axis in enumerate(self.items):
            # Tee the colors iterator so we can peek at the next color. The next color in the iterator becomes the
            # axis color. The first series on each axis should match the axis color, and will progress the colors
            # iterator. The next axis color should be the next color in the iterator after the last color used by a
            # series on the current axis in order to get a better variety of color.
            colors, tee_colors = itertools.tee(colors)
            axis_color = next(tee_colors)

            # prepend axes, append series, this keeps everything ordered left-to-right
            y_axes[0:0] = self._render_y_axis(axis_idx,
                                              axis_color if 1 < total_num_series else None,
                                              references)

            series += self._render_series(axis,
                                          axis_idx,
                                          axis_color,
                                          colors,
                                          data_frame,
                                          series_data_frames,
                                          dimension_fields,
                                          references,
                                          is_timeseries)

        x_axis = self._render_x_axis(data_frame, dimensions, slicer.fields)

        return {
            "title": {"text": self.title},
            "xAxis": x_axis,
            "yAxis": y_axes,
            'colors': self.colors,
            "series": series,
            "tooltip": {"shared": True, "useHTML": True, "enabled": self.tooltip_visible},
            "legend": {"useHTML": True},
        }

    def _render_x_axis(self, data_frame, dimensions, fields):
        """
        Renders the xAxis configuration.

        https://api.highcharts.com/highcharts/xAxis

        :param data_frame:
        :param dimensions:
        :return:
        """
        is_mi = isinstance(data_frame.index, pd.MultiIndex)
        first_level = data_frame.index.levels[0] \
            if is_mi \
            else data_frame.index

        is_timeseries = dimensions and dimensions[0].data_type == DataType.date
        if is_timeseries:
            return {
                "type": "datetime",
                "visible": self.x_axis_visible,
            }

        categories = ["All"]
        if first_level.name is not None:
            dimension_alias = utils.alias_for_alias_selector(first_level.name)
            dimension = fields[dimension_alias]
            categories = [formats.display_value(category, dimension) or category
                          for category in first_level]

        return {
            "type": "category",
            "categories": categories,
            "visible": self.x_axis_visible,
        }

    def _render_y_axis(self, axis_idx, color, references):
        """
        Renders the yAxis configuration.

        https://api.highcharts.com/highcharts/yAxis

        :param axis_idx:
        :param color:
        :param references:
        :return:
        """
        axis = self.items[axis_idx]

        y_axes = [{
            "id": str(axis_idx),
            "title": {"text": None},
            "labels": {"style": {"color": color}},
            "visible": axis.y_axis_visible,
        }]

        y_axes += [{
            "id": "{}_{}".format(axis_idx, reference.alias),
            "title": {"text": reference.label},
            "opposite": True,
            "labels": {"style": {"color": color}},
            "visible": axis.y_axis_visible,
        }
            for reference in references
            if reference.delta]

        return y_axes

    def _render_series(self, axis, axis_idx, axis_color, colors, data_frame, series_data_frames, dimension_fields,
                       references, is_timeseries=False):
        """
        Renders the series configuration.

        https://api.highcharts.com/highcharts/series

        :param axis:
        :param axis_idx:
        :param axis_color:
        :param colors:
        :param series_data_frames:
        :param dimension_fields:
        :param references:
        :param is_timeseries:
        :return:
        """

        hc_series = []
        for series in axis:
            # Pie Charts, do not break data out into groups, render all data in one pie chart series
            if isinstance(series, self.PieSeries):
                # Pie charts do not group the data up by series so render them separately and then go to the next series
                for reference in [None] + references:
                    hc_series.append(self._render_pie_series(series.metric,
                                                             reference,
                                                             data_frame,
                                                             dimension_fields))
                continue

            # For other series types, create a highcharts series for each group (combination of dimension values)

            symbols = itertools.cycle(MARKER_SYMBOLS)
            for (dimension_values, group_df), symbol in zip(series_data_frames, symbols):
                dimension_values = utils.wrap_list(dimension_values)
                dimension_label = self._format_dimension_values(dimension_fields[1:], dimension_values)

                hc_series += self._render_highcharts_series(series,
                                                            group_df,
                                                            references,
                                                            dimension_label,
                                                            is_timeseries,
                                                            symbol,
                                                            axis_idx,
                                                            axis_color,
                                                            next(colors))

        return hc_series

    def _render_highcharts_series(self, series, series_df, references, dimension_label, is_timeseries, symbol, axis_idx,
                                  axis_color, series_color):
        """

        Note on colors:
        - With a single axis, use different colors for each series
        - With multiple axes, use the same color for the entire axis and only change the dash style

        :param series:
        :param series_df:
        :param references:
        :param dimension_label:
        :param is_timeseries:
        :param symbol:
        :param axis_idx:
        :param axis_color:
        :param series_color:
        :return:
        """
        if is_timeseries:
            series_df = series_df.sort_index(level=0)

        results = []
        for reference, dash_style in zip([None] + references, itertools.cycle(DASH_STYLES)):
            field_alias = utils.alias_selector(reference_alias(series.metric, reference))
            metric_label = reference_label(series.metric, reference)

            hc_series = ({
                "type": series.type,
                "name": '{} ({})'.format(metric_label, dimension_label) if dimension_label else metric_label,

                "data": (
                    self._render_timeseries_data(series_df, field_alias, series.metric)
                    if is_timeseries
                    else self._render_category_data(series_df, field_alias, series.metric)
                ),

                "tooltip": self._render_tooltip(series.metric, reference),

                "yAxis": ("{}_{}".format(axis_idx, reference.alias)
                          if reference is not None and reference.delta
                          else str(axis_idx)),

                "marker": ({"symbol": symbol, "fillColor": axis_color or series_color}
                           if isinstance(series, SERIES_NEEDING_MARKER)
                           else {}),

                "stacking": series.stacking,
            })

            if isinstance(series, ContinuousAxisSeries):
                # Set each series in a continuous series to a specific color
                hc_series["color"] = series_color
                hc_series["dashStyle"] = dash_style

            results.append(hc_series)

        return results

    @staticmethod
    def _render_category_data(group_df, field_alias, metric):
        categories = list(group_df.index.levels[0]) \
            if isinstance(group_df.index, pd.MultiIndex) \
            else list(group_df.index)

        series = []
        for labels, y in group_df[field_alias].iteritems():
            label = labels[0] if isinstance(labels, tuple) else labels

            series.append({
                'x': categories.index(label),
                'y': formats.raw_value(y, metric)
            })

        return series

    @staticmethod
    def _render_timeseries_data(group_df, metric_alias, metric):
        series = []
        for dimension_values, y in group_df[metric_alias].iteritems():
            first_dimension_value = utils.wrap_list(dimension_values)[0]

            # Ignore empty result sets where the only row is totals
            if first_dimension_value in TOTALS_MARKERS:
                continue

            if pd.isnull(first_dimension_value):
                # Ignore totals on the x-axis.
                continue

            series.append((formats.date_as_millis(first_dimension_value),
                           formats.raw_value(y, metric)))
        return series

    def _render_tooltip(self, metric, reference):
        return {
            "valuePrefix": reference_prefix(metric, reference),
            "valueSuffix": reference_suffix(metric, reference),
            "valueDecimals": metric.precision,
        }

    def _render_pie_series(self, metric, reference, data_frame, dimension_fields):
        metric_alias = utils.alias_selector(metric.alias)

        data = []
        for dimension_values, y in data_frame[metric_alias].iteritems():
            dimension_values = utils.wrap_list(dimension_values)
            name = self._format_dimension_values(dimension_fields, dimension_values)

            data.append({
                "name": name or metric.label,
                "y": formats.raw_value(y, metric),
            })

        return {
            "name": reference_label(metric, reference),
            "type": 'pie',
            "data": data,
            'tooltip': {
                'pointFormat': '<span style="color:{point.color}">\u25CF</span> {series.name}: '
                               '<b>{point.y} ({point.percentage:.1f}%)</b><br/>',
                'valueDecimals': metric.precision,
                'valuePrefix': reference_prefix(metric, reference),
                'valueSuffix': reference_suffix(metric, reference),
            },
        }

    @staticmethod
    def _remove_date_totals(data_frame):
        """
        This function filters the totals value for the date/time dimension from the result set. There is no way to
        represent this value on a chart so it is just removed.

        :param data_frame:
        :return:
        """
        if isinstance(data_frame.index, pd.MultiIndex):
            index_slice = data_frame.index.get_level_values(0) < TS_UPPER_BOUND
            return data_frame.loc[index_slice, :]

        if isinstance(data_frame.index, pd.DatetimeIndex):
            return data_frame[data_frame.index < TS_UPPER_BOUND]

        return data_frame

    @staticmethod
    def _group_by_series(data_frame):
        if len(data_frame) == 0 or not isinstance(data_frame.index, pd.MultiIndex):
            return [([], data_frame)]

        levels = data_frame.index.names[1:]
        return data_frame.groupby(level=levels, sort=False)

    @staticmethod
    def _format_dimension_values(dimension_fields, dimension_values):
        return ', '.join(str.strip(formats.display_value(value, dimension_field) or str(value))
                         for value, dimension_field in zip(dimension_values, dimension_fields))



import copy
from collections import (
    defaultdict,
    namedtuple,
)

from toposort import (
    CircularDependencyError,
    toposort_flatten,
)

from fireant.dataset.intervals import (
    DATETIME_INTERVALS,
    DatetimeInterval,
)
from fireant.dataset.modifiers import (
    OmitFromRollup,
    Rollup,
)
from fireant.dataset.operations import Share
from fireant.exceptions import SlicerException
from fireant.utils import (
    groupby,
    ordered_distinct_list,
    ordered_distinct_list_by_attr,
)


class MissingTableJoinException(SlicerException):
    pass


class CircularJoinsException(SlicerException):
    pass


ReferenceGroup = namedtuple('ReferenceGroup', ('dimension', 'time_unit', 'intervals'))


def find_required_tables_to_join(elements, base_table):
    """
    Collect all the tables required for a given list of slicer elements.  This looks through the definition and
    display_definition attributes of all elements and

    This looks through the metrics, dimensions, and filter included in this slicer query. It also checks both the
    definition
    field of each element as well as the display definition for Unique Dimensions.

    :return:
        A collection of tables required to execute a query,
    """
    return ordered_distinct_list([table
                                  for element in elements

                                  # Need extra for-loop to incl. the `display_definition` from `UniqueDimension`
                                  for attr in [getattr(element, 'definition', None)]

                                  # ... but then filter Nones since most elements do not have `display_definition`
                                  if attr is not None

                                  for table in attr.tables_

                                  # Omit the base table from this list
                                  if base_table != table])


def find_joins_for_tables(joins, base_table, required_tables):
    """
    Given a set of tables required for a slicer query, this function finds the joins required for the query and
    sorts them topologically.

    :return:
        A list of joins in the order that they must be joined to the query.
    :raises:
        MissingTableJoinException - If a table is required but there is no join for that table
        CircularJoinsException - If there is a circular dependency between two or more joins
    """
    dependencies = defaultdict(set)
    slicer_joins = {join.table: join
                    for join in joins}

    while required_tables:
        table = required_tables.pop()

        if table not in slicer_joins:
            raise MissingTableJoinException('Could not find a join for table {}'
                                            .format(str(table)))

        join = slicer_joins[table]
        tables_required_for_join = set(join.criterion.tables_) - {base_table, join.table}

        dependencies[join] |= {slicer_joins[table]
                               for table in tables_required_for_join}
        required_tables += tables_required_for_join - {d.table for d in dependencies}

    try:
        return toposort_flatten(dependencies, sort=True)
    except CircularDependencyError as e:
        raise CircularJoinsException(str(e))


def find_metrics_for_widgets(widgets):
    """
    :return:
        an ordered, distinct list of metrics used in all widgets as part of this query.
    """
    return ordered_distinct_list_by_attr([metric
                                          for widget in widgets
                                          for metric in widget.metrics])


def find_share_dimensions(dimensions, operations):
    """
    Returns a subset list of dimensions from the list of dimensions that are used as the over-dimension in share
    operations.

    :param dimensions:
    :param operations:
    :return:
    """
    share_operations_over_dimensions = [operation.over
                                        for operation in operations
                                        if isinstance(operation, Share)
                                        and operation.over is not None]

    dimension_map = {dimension.alias: dimension
                     for dimension in dimensions}

    return [dimension_map[dimension.alias]
            for dimension in share_operations_over_dimensions]


def find_operations_for_widgets(widgets):
    """
    :return:
        an ordered, distinct list of metrics used in all widgets as part of this query.
    """
    return ordered_distinct_list_by_attr([operation
                                          for widget in widgets
                                          for operation in widget.operations])


def find_totals_dimensions(dimensions, share_dimensions):
    """
    :param dimensions:
    :param share_dimensions:
    :return:
        an list of all dimension field in the list argument `dimensions` which have the `Rollup` modifier applied to
        them or are used as a basis for a share metric.
    """
    share_dimension_aliases = {d.alias for d in share_dimensions}
    return [dimension
            for dimension in dimensions
            if isinstance(dimension, Rollup)
            or dimension.alias in share_dimension_aliases]


def find_filters_for_totals(filters):
    """
    :param filters:
    :return:
        a list of filters that should be applied to totals queries. This removes any filters from the list that have the
        `OmitFromRollup` modifier applied to them.
    """
    return [fltr
            for fltr in filters
            if not isinstance(fltr, OmitFromRollup)
            and not fltr.is_aggregate]


def find_and_replace_reference_dimensions(references, dimensions):
    """
    Finds the dimension for a reference in the query if there is one and replaces it. This is to force the reference to
    use the same modifiers with a dimension if it is selected in the query.

    :param references:
    :param dimensions:
    :return:
    """
    dimensions_by_key = {dimension.alias: dimension
                         for dimension in dimensions}

    reference_copies = []
    for reference in map(copy.deepcopy, references):
        dimension = dimensions_by_key.get(reference.field.alias)
        if dimension is not None:
            reference.field = dimension
        reference_copies.append(reference)
    return reference_copies


interval_weekdays = {
    'month': ('week', 4),
    'quarter': ('week', 4 * 3),
    'year': ('week', 4 * 13),
}


def find_and_group_references_for_dimensions(dimensions, references):
    """
    Finds all of the references for dimensions and groups them by dimension, interval unit, number of intervals.

    This structure reflects how the references need to be joined to the slicer query. References of the same
    type (WoW, WoW.delta, WoW.delta_percent) can share a join query.

    :param dimensions:
    :param references:

    :return:
        An `OrderedDict` where the keys are 3-item tuples consisting of "Dimension, interval unit, # of intervals.

        .. code-block:: python

            Example
            {
                (Dimension(date_1), 'weeks', 1): [WoW, WoW.delta],
                (Dimension(date_1), 'years', 1): [YoY],
                (Dimension(date_7), 'days', 1): [DoD, DoD.delta_percent],
            }
    """
    align_weekdays = dimensions \
                     and isinstance(dimensions[0], DatetimeInterval) \
                     and -1 < DATETIME_INTERVALS.index(dimensions[0].interval_key) < 3

    def get_dimension_time_unit_and_interval(reference):
        defaults = (reference.time_unit, 1)
        time_unit, interval_muliplier = interval_weekdays.get(reference.time_unit, defaults) \
            if align_weekdays \
            else defaults

        return reference.field, time_unit, interval_muliplier * reference.interval

    distinct_references = ordered_distinct_list(references)
    return groupby(distinct_references, get_dimension_time_unit_and_interval)



from typing import (
    Dict,
    Iterable,
)

import pandas as pd

from fireant.dataset.fields import Field
from fireant.dataset.totals import scrub_totals_from_share_results
from fireant.exceptions import SlicerException
from fireant.reference_helpers import reference_alias
from fireant.utils import (
    alias_for_alias_selector,
    alias_selector,
    immutable,
)
from pypika import Order
from . import special_cases
from .execution import fetch_data
from .field_helper import make_orders_for_dimensions
from .finders import (
    find_and_group_references_for_dimensions,
    find_and_replace_reference_dimensions,
    find_metrics_for_widgets,
    find_operations_for_widgets,
    find_share_dimensions,
)
from .pagination import paginate
from .sql_transformer import (
    make_latest_query,
    make_slicer_query,
    make_slicer_query_with_totals_and_references,
)
from ..formats import display_value


class QueryException(SlicerException):
    pass


def add_hints(queries, hint=None):
    return [query.hint(hint)
            if hint is not None and hasattr(query.__class__, 'hint')
            else query
            for query in queries]


class QueryBuilder(object):
    """
    This is the base class for building slicer queries. This class provides an interface for building slicer queries
    via a set of functions which can be chained together.
    """

    def __init__(self, dataset, table):
        self.dataset = dataset
        self.table = table
        self._dimensions = []
        self._filters = []
        self._apply_filter_to_totals = []
        self._references = []
        self._limit = None
        self._offset = None

    @immutable
    def filter(self, *filters, apply_to_totals=True):
        """
        :param filters:
        :param apply_to_totals:
        :return:
        """
        self._filters += [f for f in filters]
        self._apply_filter_to_totals += [apply_to_totals] * len(filters)

    @immutable
    def limit(self, limit):
        """
        :param limit:
            A limit on the number of database rows returned.
        """
        self._limit = limit

    @immutable
    def offset(self, offset):
        """
        :param offset:
            A offset on the number of database rows returned.
        """
        self._offset = offset

    @property
    def sql(self):
        """
        Serialize this query builder object to a set of Pypika/SQL queries.

        This is the base implementation shared by two implementations: the query to fetch data for a slicer request and
        the query to fetch choices for dimensions.

        This function only handles dimensions (select+group by) and filtering (where/having), which is everything needed
        for the query to fetch choices for dimensions.

        The slicer query extends this with metrics, references, and totals.
        """
        raise NotImplementedError()

    def fetch(self, hint=None):
        """
        Fetches the data for this query instance and returns it in an instance of `pd.DataFrame`

        :param hint:
            For database vendors that support it, add a query hint to collect analytics on the queries triggerd by
            fireant.
        """
        queries = add_hints(self.sql, hint)

        return fetch_data(self.dataset.database, queries, self._dimensions)


class DataSetQueryBuilder(QueryBuilder):
    """
    Slicer queries consist of widgets, dimensions, filters, and references. At least one or more widgets is required.
    All others are optional.
    """

    def __init__(self, dataset):
        super(DataSetQueryBuilder, self).__init__(dataset, dataset.table)
        self._widgets = []
        self._totals_dimensions = set()
        self._orders = []

    @immutable
    def widget(self, *widgets):
        """

        :param widgets:
        :return:
        """
        self._widgets += widgets

    @immutable
    def dimension(self, *dimensions):
        """
        :param dimensions:
        :return:
        """
        aliases = {dimension.alias
                   for dimension in self._dimensions}
        self._dimensions += [dimension
                             for dimension in dimensions
                             if dimension.alias not in aliases]

    @immutable
    def reference(self, *references):
        """
        Add a reference for a dimension when building a slicer query.

        :param references:
            References to add to the query
        :return:
            A copy of the dimension with the reference added.
        """
        self._references += references

    @immutable
    def orderby(self, field: Field, orientation: Order = None):
        """
        :param field:
            The element to order by, either a metric or dimension.
        :param orientation:
            The directionality to order by, either ascending or descending.
        :return:
        """
        self._orders += [(field.definition.as_(alias_selector(field.alias)), orientation)]

    def _validate(self):
        for widget in self._widgets:
            if hasattr(widget, 'validate'):
                widget.validate(self._dimensions)

    @property
    def reference_groups(self):
        return list(find_and_group_references_for_dimensions(self._dimensions, self._references).values())

    @property
    def sql(self):
        """
        Serialize this query builder to a list of Pypika/SQL queries. This function will return one query for every
        combination of reference and rolled up dimension (including null options).

        This collects all of the metrics in each widget, dimensions, and filters and builds a corresponding pypika query
        to fetch the data.  When references are used, the base query normally produced is wrapped in an outer query and
        a query for each reference is joined based on the referenced dimension shifted.
        """
        # First run validation for the query on all widgets
        self._validate()

        metrics = find_metrics_for_widgets(self._widgets)
        operations = find_operations_for_widgets(self._widgets)
        share_dimensions = find_share_dimensions(self._dimensions, operations)
        references = find_and_replace_reference_dimensions(self._references, self._dimensions)
        orders = (self._orders or make_orders_for_dimensions(self._dimensions))

        return make_slicer_query_with_totals_and_references(self.dataset.database,
                                                            self.table,
                                                            self.dataset.joins,
                                                            self._dimensions,
                                                            metrics,
                                                            operations,
                                                            self._filters,
                                                            references,
                                                            orders,
                                                            share_dimensions=share_dimensions)

    def fetch(self, hint=None) -> Iterable[Dict]:
        """
        Fetch the data for this query and transform it into the widgets.

        :param hint:
            A query hint label used with database vendors which support it. Adds a label comment to the query.
        :return:
            A list of dict (JSON) objects containing the widget configurations.
        """
        queries = add_hints(self.sql, hint)

        operations = find_operations_for_widgets(self._widgets)
        share_dimensions = find_share_dimensions(self._dimensions, operations)

        data_frame = fetch_data(self.dataset.database,
                                queries,
                                self._dimensions,
                                share_dimensions,
                                self.reference_groups)

        # Apply operations
        for operation in operations:
            for reference in [None] + self._references:
                df_key = alias_selector(reference_alias(operation, reference))
                data_frame[df_key] = operation.apply(data_frame, reference)

        data_frame = scrub_totals_from_share_results(data_frame, self._dimensions)
        data_frame = special_cases.apply_operations_to_data_frame(operations, data_frame)
        data_frame = paginate(data_frame,
                              self._widgets,
                              orders=self._orders,
                              limit=self._limit,
                              offset=self._offset)

        # Apply transformations
        return [widget.transform(data_frame, self.dataset, self._dimensions, self._references)
                for widget in self._widgets]

    def __str__(self):
        return str(self.sql)

    def __repr__(self):
        return ".".join(["slicer", "data"]
                        + ["widget({})".format(repr(widget))
                           for widget in self._widgets]
                        + ["dimension({})".format(repr(dimension))
                           for dimension in self._dimensions]
                        + ["filter({}{})".format(repr(f),
                                                 ', apply_filter_to_totals=True' if apply_filter_to_totals else '')
                           for f, apply_filter_to_totals in zip(self._filters, self._apply_filter_to_totals)]
                        + ["reference({})".format(repr(reference))
                           for reference in self._references]
                        + ["orderby({}, {})".format(definition.alias,
                                                    orientation)
                           for (definition, orientation) in self._orders])


class DimensionChoicesQueryBuilder(QueryBuilder):
    """
    This builder is used for building slicer queries for fetching the choices for a dimension given a set of filters.
    """

    def __init__(self, dataset, dimension):
        super(DimensionChoicesQueryBuilder, self).__init__(dataset, dataset.hint_table or dataset.table)
        self._dimensions.append(dimension)

        # TODO remove after 3.0.0
        display_alias = dimension.alias + '_display'
        if display_alias in dataset.fields:
            self._dimensions.append(dataset.fields[display_alias])

    @property
    def sql(self):
        """
        Serializes this query builder as a set of SQL queries.  This method will always return a list of one query since
        only one query is required to retrieve dimension choices.

        This function only handles dimensions (select+group by) and filtering (where/having), which is everything needed
        for the query to fetch choices for dimensions.

        The slicer query extends this with metrics, references, and totals.
        """
        query = make_slicer_query(database=self.dataset.database,
                                  base_table=self.table,
                                  joins=self.dataset.joins,
                                  dimensions=self._dimensions,
                                  filters=self._filters) \
            .limit(self._limit) \
            .offset(self._offset)

        return [query]

    def fetch(self, hint=None, force_include=()) -> pd.Series:
        """
        Fetch the data for this query and transform it into the widgets.

        :param hint:
            For database vendors that support it, add a query hint to collect analytics on the queries triggerd by
            fireant.
        :param force_include:
            A list of dimension values to include in the result set. This can be used to avoid having necessary results
            cut off due to the pagination.  These results will be returned at the head of the results.
        :return:
            A list of dict (JSON) objects containing the widget configurations.
        """
        query = add_hints(self.sql, hint)[0]

        dimension = self._dimensions[0]
        definition = dimension.definition.as_(alias_selector(dimension.alias))

        if force_include:
            include = self.dataset.database.to_char(dimension.definition) \
                .isin([str(x) for x in force_include])

            # Ensure that these values are included
            query = query.orderby(include, order=Order.desc)

        # Order by the dimension definition that the choices are for
        query = query.orderby(definition)

        data = fetch_data(self.dataset.database, [query], self._dimensions)

        if len(data.index.names) > 1:
            display_alias = data.index.names[1]
            data.reset_index(display_alias, inplace=True)
            choices = data[display_alias]

        else:
            data['display'] = data.index.tolist()
            choices = data['display']

        dimension_display = self._dimensions[-1]
        return choices.map(lambda raw: display_value(raw, dimension_display) or raw)

    def __repr__(self):
        return ".".join(["slicer", self._dimensions[0].alias, "choices"]
                        + ["filter({})".format(repr(f))
                           for f in self._filters])


class DimensionLatestQueryBuilder(QueryBuilder):
    def __init__(self, dataset):
        super(DimensionLatestQueryBuilder, self).__init__(dataset, dataset.hint_table or dataset.table)

    @immutable
    def __call__(self, dimension: Field, *dimensions: Field):
        self._dimensions += [dimension] + list(dimensions)

    @property
    def sql(self):
        """
        Serializes this query builder as a set of SQL queries.  This method will always return a list of one
        query since
        only one query is required to retrieve dimension choices.

        This function only handles dimensions (select+group by) and filtering (where/having), which is everything
        needed
        for the query to fetch choices for dimensions.

        The slicer query extends this with metrics, references, and totals.
        """
        if not self._dimensions:
            raise QueryException('Must select at least one dimension to query latest values')

        query = make_latest_query(database=self.dataset.database,
                                  base_table=self.table,
                                  joins=self.dataset.joins,
                                  dimensions=self._dimensions)
        return [query]

    def fetch(self, hint=None):
        data = super().fetch(hint=hint).reset_index().iloc[0]
        # Remove the row index as the name and trim the special dimension key characters from the dimension key
        data.name = None
        data.index = [alias_for_alias_selector(alias)
                      for alias in data.index]
        return data
